@@ -1,16 +1,16 @@
 # This file is part of parallel-ssh.
-
-# Copyright (C) 2014-2017 Panos Kittenis
-
+#
+# Copyright (C) 2014-2022 Panos Kittenis and contributors.
+#
 # This library is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
 # License as published by the Free Software Foundation, version 2.1.
-
+#
 # This library is distributed in the hope that it will be useful,
 # but WITHOUT ANY WARRANTY; without even the implied warranty of
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 # Lesser General Public License for more details.
-
+#
 # You should have received a copy of the GNU Lesser General Public
 # License along with this library; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
@@ -20,27 +20,20 @@
 Run SSH commands over many - hundreds/hundreds of thousands - number of servers
 asynchronously and with minimal system load on the client host.
 
-New users should start with
-:py:func:`pssh.pssh_client.ParallelSSHClient.run_command`
+New users should start with `pssh.clients.ParallelSSHClient.run_command` and
+`pssh.clients.SSHClient.run_command`
 
-See also :py:class:`pssh.ParallelSSHClient` and :py:class:mod:`pssh.SSHClient`
+See also `pssh.clients.ParallelSSHClient` and pssh.clients.SSHClient`
 for class documentation.
 """
 
-# flake8: noqa: E402, F401, F402
 
-import logging
-from ._version import get_versions
-__version__ = get_versions()['version']
-del get_versions
-from .pssh_client import ParallelSSHClient
-from .ssh_client import SSHClient
-from .utils import enable_host_logger
-from .exceptions import UnknownHostException, \
-     AuthenticationException, ConnectionErrorException, SSHException
+from logging import getLogger, NullHandler
+from . import _version
+__version__ = _version.get_versions()['version']
+del _version
 
-host_logger = logging.getLogger('pssh.host_logger')
-logger = logging.getLogger('pssh')
-if hasattr(logging, 'NullHandler'):
-    host_logger.addHandler(logging.NullHandler())
-    logger.addHandler(logging.NullHandler())
+host_logger = getLogger('pssh.host_logger')
+logger = getLogger('pssh')
+host_logger.addHandler(NullHandler())
+logger.addHandler(NullHandler())
