@@ -19,10 +19,13 @@ import logging
 import os
 from getpass import getuser
 from socket import gaierror as sock_gaierror, error as sock_error
+from warnings import warn
 
 from gevent import sleep, socket, Timeout as GTimeout
 from gevent.hub import Hub
 from gevent.select import poll, POLLIN, POLLOUT
+from gevent.socket import SHUT_RDWR
+from gevent.pool import Pool
 from ssh2.exceptions import AgentConnectionError, AgentListIdentitiesError, \
     AgentAuthenticationError, AgentGetIdentityError
 from ssh2.utils import find_eol
@@ -39,6 +42,61 @@ host_logger = logging.getLogger('pssh.host_logger')
 logger = logging.getLogger(__name__)
 
 
+class PollMixIn(object):
+    """MixIn for co-operative socket polling functionality.
+
+    """
+    __slots__ = ('sock',)
+
+    def __init__(self, sock=None):
+        self.sock = sock
+
+    def poll(self, timeout=None):
+        raise NotImplementedError
+
+    def eagain(self, func, *args, **kwargs):
+        raise NotImplementedError
+
+    def eagain_write(self, write_func, data):
+        raise NotImplementedError
+
+    def _poll_errcodes(self, directions_func, inbound, outbound):
+        directions = directions_func()
+        if directions == 0:
+            return
+        events = 0
+        if directions & inbound:
+            events = POLLIN
+        if directions & outbound:
+            events |= POLLOUT
+        self._poll_socket(events)
+
+    def _poll_socket(self, events):
+        if self.sock is None:
+            return
+        poller = poll()
+        poller.register(self.sock, eventmask=events)
+        poller.poll(timeout=1)
+
+    def _eagain_errcode(self, func, eagain, *args, **kwargs):
+        ret = func(*args, **kwargs)
+        while ret == eagain:
+            self.poll()
+            ret = func(*args, **kwargs)
+            sleep()
+        return ret
+
+    def _eagain_write_errcode(self, write_func, data, eagain):
+        data_len = len(data)
+        total_written = 0
+        while total_written < data_len:
+            rc, bytes_written = write_func(data[total_written:])
+            total_written += bytes_written
+            if rc == eagain:
+                self.poll()
+            sleep()
+
+
 class Stdin(object):
     """Stdin stream for a channel.
 
@@ -64,11 +122,13 @@ class Stdin(object):
         :param data: Data to write.
         :type data: str
         """
-        return self._client._eagain(self._channel.write, data)
+        return self._client.eagain(self._channel.write, data)
 
     def flush(self):
         """Flush pending data written to stdin."""
-        return self._client._eagain(self._channel.flush)
+        if not hasattr(self._channel, "flush"):
+            return
+        return self._client.eagain(self._channel.flush)
 
 
 class InteractiveShell(object):
@@ -130,7 +190,7 @@ class InteractiveShell(object):
         """Wait for shell to finish executing and close channel."""
         if self._chan is None:
             return
-        self._client._eagain(self._chan.send_eof)
+        self._client.eagain(self._chan.send_eof)
         self._client.wait_finished(self.output)
         return self
 
@@ -142,10 +202,10 @@ class InteractiveShell(object):
         :type cmd: str
         """
         cmd = cmd.encode(self._encoding) + self._EOL
-        self._client._eagain_write(self._chan.write, cmd)
+        self._client.eagain_write(self._chan.write, cmd)
 
 
-class BaseSSHClient(object):
+class BaseSSHClient(PollMixIn):
 
     IDENTITIES = (
         os.path.expanduser('~/.ssh/id_rsa'),
@@ -167,6 +227,7 @@ class BaseSSHClient(object):
                  identity_auth=True,
                  ipv6_only=False,
                  ):
+        super(PollMixIn, self).__init__()
         self._auth_thread_pool = _auth_thread_pool
         self.host = host
         self.alias = alias
@@ -174,7 +235,6 @@ class BaseSSHClient(object):
         self.password = password
         self.port = port if port else 22
         self.num_retries = num_retries
-        self.sock = None
         self.timeout = timeout if timeout else None
         self.retry_delay = retry_delay
         self.allow_agent = allow_agent
@@ -185,6 +245,7 @@ class BaseSSHClient(object):
         self.identity_auth = identity_auth
         self._keepalive_greenlet = None
         self.ipv6_only = ipv6_only
+        self._pool = Pool()
         self._init()
 
     def _pkey_from_memory(self, pkey_data):
@@ -210,11 +271,15 @@ class BaseSSHClient(object):
             raise AuthenticationError(msg, self.host, self.port, ex, retries, self.num_retries)
 
     def disconnect(self):
+        """Deprecated and a no-op. Disconnections handled by client de-allocation."""
+        warn("Deprecated and a no-op - to be removed in future releases.", DeprecationWarning)
+
+    def _disconnect(self):
         raise NotImplementedError
 
     def __del__(self):
         try:
-            self.disconnect()
+            self._disconnect()
         except Exception:
             pass
 
@@ -222,7 +287,7 @@ class BaseSSHClient(object):
         return self
 
     def __exit__(self, *args):
-        self.disconnect()
+        self._disconnect()
 
     def open_shell(self, encoding='utf-8', read_timeout=None):
         """Open interactive shell on new channel.
@@ -242,7 +307,8 @@ class BaseSSHClient(object):
         raise NotImplementedError
 
     def _disconnect_eagain(self):
-        self._eagain(self.session.disconnect)
+        if self.session is not None and self.sock is not None and not self.sock.closed:
+            self.eagain(self.session.disconnect)
 
     def _connect_init_session_retry(self, retries):
         try:
@@ -252,9 +318,11 @@ class BaseSSHClient(object):
         self.session = None
         if not self.sock.closed:
             try:
-                self.sock.close()
+                self.sock.shutdown(SHUT_RDWR)
+                self.sock.detach()
             except Exception:
                 pass
+        self.sock = None
         sleep(self.retry_delay)
         self._connect(self._host, self._port, retries=retries)
         return self._init_session(retries=retries)
@@ -418,6 +486,13 @@ class BaseSSHClient(object):
         raise NotImplementedError
 
     def execute(self, cmd, use_pty=False, channel=None):
+        """
+        Deprecated - use ``run_command`` instead which returns a ``HostOutput`` object.
+        """
+        warn("Deprecated - use run_command instead.", DeprecationWarning)
+        return self._execute(cmd, use_pty=use_pty, channel=channel)
+
+    def _execute(self, cmd, use_pty=False, channel=None):
         raise NotImplementedError
 
     def read_stderr(self, stderr_buffer, timeout=None):
@@ -552,37 +627,11 @@ class BaseSSHClient(object):
             _command += "%s '%s'" % (_shell, command,)
         _command = _command.encode(encoding)
         with GTimeout(seconds=self.timeout):
-            channel = self.execute(_command, use_pty=use_pty)
+            channel = self._execute(_command, use_pty=use_pty)
         _timeout = read_timeout if read_timeout else timeout
         host_out = self._make_host_output(channel, encoding, _timeout)
         return host_out
 
-    def _eagain_write_errcode(self, write_func, data, eagain):
-        data_len = len(data)
-        total_written = 0
-        while total_written < data_len:
-            rc, bytes_written = write_func(data[total_written:])
-            total_written += bytes_written
-            if rc == eagain:
-                self.poll()
-            sleep()
-
-    def _eagain_errcode(self, func, eagain, *args, **kwargs):
-        timeout = kwargs.pop('timeout', self.timeout)
-        with GTimeout(seconds=timeout, exception=Timeout):
-            ret = func(*args, **kwargs)
-            while ret == eagain:
-                self.poll()
-                ret = func(*args, **kwargs)
-                sleep()
-            return ret
-
-    def _eagain_write(self, write_func, data):
-        raise NotImplementedError
-
-    def _eagain(self, func, *args, **kwargs):
-        raise NotImplementedError
-
     def _make_sftp(self):
         raise NotImplementedError
 
@@ -690,24 +739,3 @@ class BaseSSHClient(object):
         _sep = file_path.rfind('/')
         if _sep > 0:
             return file_path[:_sep]
-
-    def poll(self):
-        raise NotImplementedError
-
-    def _poll_socket(self, events):
-        if self.sock is None:
-            return
-        poller = poll()
-        poller.register(self.sock, eventmask=events)
-        poller.poll(timeout=1)
-
-    def _poll_errcodes(self, directions_func, inbound, outbound):
-        directions = directions_func()
-        if directions == 0:
-            return
-        events = 0
-        if directions & inbound:
-            events = POLLIN
-        if directions & outbound:
-            events |= POLLOUT
-        self._poll_socket(events)
