@@ -1,20 +1,28 @@
 # This file is part of parallel-ssh.
-
-# Copyright (C) 2014-2017 Panos Kittenis
-
+#
+# Copyright (C) 2014-2022 Panos Kittenis and contributors.
+#
 # This library is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
 # License as published by the Free Software Foundation, version 2.1.
-
+#
 # This library is distributed in the hope that it will be useful,
 # but WITHOUT ANY WARRANTY; without even the implied warranty of
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 # Lesser General Public License for more details.
-
+#
 # You should have received a copy of the GNU Lesser General Public
 # License along with this library; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
+
 """Constants definitions for pssh package"""
 
+
 DEFAULT_RETRIES = 3
+"""Default number of retry attempts for SSH client initialisation - authentication, establishing
+connections et al."""
+
+
+RETRY_DELAY = 5
+"""Default delay in seconds between retry attempts for SSH client initialisation."""
